Require Import Le Plus List Coq.Arith.Compare Coq.Program.Tactics.
Require Import ILM ILM_Types.

Open Scope map_scope.

Set Implicit Arguments.

(** Example(s) for well-typed program(s) *)
(* This also ensures such programs exist... *)
Section Examples.
  Parameter e:exp.
  Parameter t:ty.
  Parameter v:val.
  Hypothesis np:~isPtr t.
  Hypothesis et:expOfType ompty e t.
  Hypothesis vt:valOfType v t.

  Definition a  := mkvar 0.
  Definition x  := mkvar 1.
  Definition x1 := mkvar 2.
  Definition x2 := mkvar 3.
  Definition a1 := mkvar 4.
  Definition a2 := mkvar 5.

  Definition m  := mkvar 0.
  Definition m' := mkvar 1.
  Definition m1 := mkvar 2.
  Definition m2 := mkvar 3.

  Definition ea := @BSet.empty alias_token.
  Hint Unfold ea.
  Lemma ea_applies (E:env val): alias_tokens_apply E ea.
  Proof.
     intros ? IN. crush.
  Qed.

  Definition example1 := stmtExp    ea x  e  (
                         stmtReturn ea x ).
  Lemma example1_types: stmtOfType (empty_env m') example1 t.
  Proof.
    eapply WT_Exp, WT_Return; crush.
  Qed.

  Definition example2 := stmtAlloc  ea               a1 m1 (
                         stmtAlloc  ea               a2 m2 (
                         stmtStore  {{AliasNeq a1 a2}} m1 m1 a1 x1 (
                         stmtStore  {{AliasNeq a1 a2}} m2 m2 a2 x2 (
                         stmtMerge  {{AliasNeq a1 a2}} m  m1 m2 (
                         stmtLoad   {{AliasNeq a1 a2}} x  m  a1 (
                         stmtReturn {{AliasNeq a1 a2}} x )))))).
  Definition example2_te:type_env :=
    (ompty[x1 <- Some t][x2 <- Some t], nil, ompty[m' <- Some MemPanorama], BSet.empty).
  Lemma example2_types: stmtOfType example2_te example2 t.
  Proof.
    eapply WT_Alloc, WT_Alloc, WT_Store, WT_Store, WT_MergeF, WT_Load, WT_Return; try crush.
    - intros a' IN. rewrite empty_neutral_union, element_single in IN. subst a'.
      apply AliasNeqSym, AliasAss. simpl. cset_tac. auto.
    - simpl. exists a1. split; cset_tac; auto using AliasEqRefl.
    - simpl. exists a2. split; cset_tac; auto using AliasEqRefl.
    - simpl. exists a1. split; cset_tac; auto using AliasEqRefl.
  Qed.
  Lemma example2_state_types: stateConsistent example2_te (@BMap.empty val (option ty) None)
                                   (empty[x1 <- v][x2 <- v], nil, empty[m' <- mempty], BSet.empty, example2) t.
  Proof.
    split.
    - exact example2_types.
    - intros x t HASTY. crush.
    - intros n ET ZT rt' GET. now inversion GET.
    - eapply memPanoramaOnlyTypes with (mv := mempty).
      + intros x IN. exfalso. apply domain_spec2 in IN. crush.
      + crush.
      + crush.
    - split.
      + crush.
      + intros a' t' EQ. exfalso. assert (t <> ptrTo t'). { intros EQ2. apply np. rewrite EQ2. apply ptrIsPtr. }
        crush.
      + crush.
      + crush.
  Qed.
  Lemma example2_alias_correct: alias_tokens_correct example2_te example2 t.
  Proof.
    intros ? ? ? ? ? _ ? ? ? steps EQ1 EQ2. unfold example2, example2_te, state_G in *.
    invc steps; simpl; [crush|]. invc H. invc H0; simpl; [crush|].
    assert (NOTINU:forall U x, x <> allocator (U ∪ {{x}})).
    { clear. intros U x EQ. apply allocatorGivesNew with (used:=(U ∪ {{x}})). rewrite <-EQ. cset_tac. auto. }
    assert (TAP:alias_tokens_apply (E [a1 <- allocator U] [a2 <- allocator (U ∪ {{allocator U}})])
                               {{AliasNeq a1 a2}}).
    { intros ? IN. rewrite element_single in IN. subst t0. simpl. assert (a1 <> a2) by crush. simplify lookup'.
      eapply NOTINU. }
    invc H. invc H1; simpl in *; [assumption|]. invc H. invc H0; simpl; [assumption|].
    invc H. invc H1; simpl; [assumption|]. invc H. invc H0; simpl; [assumption|]. clear H11.
    invc H. invc H1; simpl.
    - intros ? IN. rewrite element_single in IN. subst t0. simpl. assert (a1 <> a2) by crush.
      assert (a1 <> x) by crush. assert (x <> a2) by crush. simplify lookup'. eapply NOTINU.
    - invc H.
  Qed.

  Definition example3 := stmtExp ea x1 e (
                         stmtExp ea x2 e example2).
  Lemma example3_types: stmtOfType (empty_env m') example3 t.
  Proof.
    eapply WT_Exp, WT_Exp, example2_types; try crush.
    - eapply expOfType_subEnv, et. intros x y VAL. crush.
  Qed.
End Examples.

(* 
*** Local Variables: ***
*** coq-load-path: ("infra" "constr" "il" "isa" ".") ***
*** End: ***
*)
