Require Import Util EqDec CSet Map.
Import BSet BMap.
Set Implicit Arguments.

(** * More or less abstract values *)
(** A file that abstacts over values and (should) provide all necessary operations. Although we concretely instantiate val to int, we maintain this file as interface between our proofs and Integer library. *)

Variable val : Type.
Variable default_val : val.

Global Instance inst_val_defaulted : Defaulted val := {
  default_el := default_val
}.

Hypothesis val_eq_dec_p : EqDec val eq.
Global Instance val_eq_dec:EqDec val eq := val_eq_dec_p.
(** ** There must be an injection into the booleans *)
Hypothesis val2bool : val -> bool.

Hypothesis ty : Type.

Hypothesis inst_eq_dec_ty_p : EqDec ty eq.
Global Instance inst_eq_dec_ty : EqDec ty eq := inst_eq_dec_ty_p.


Hypothesis valOfType : val -> ty -> Prop.
Hypothesis valSynth : forall t:ty, { v:val | valOfType v t }.

Parameter ptrTo : ty -> ty.
Hypothesis isPtr_sigma: forall t:ty, { pt | t = ptrTo pt } + { forall pt, ptrTo pt <> t }.
Hypothesis ptrTo_injective: forall t1 t2:ty, ptrTo t1 = ptrTo t2 -> t1 = t2.

Hypothesis allocator : cset val -> val. (* used locations -> new location *)
Hypothesis allocatorGivesPtr : forall used, forall ty, valOfType (allocator used) (ptrTo ty).
Hypothesis allocatorGivesNew : forall used, allocator used ∉ used.

(* 
*** Local Variables: ***
*** coq-load-path: ("../infra" "../constr") ***
*** End: ***
*)
